@@ -269,7 +269,7 @@ DEFINE_TEST(test_entry)
 	archive_entry_set_hardlink(e, "hardlink");
 	archive_entry_set_symlink(e, "symlink");
 	archive_entry_set_link(e, "link");
-	assertEqualString(archive_entry_hardlink(e), "hardlink");
+	assertEqualString(archive_entry_hardlink(e), NULL);
 	assertEqualString(archive_entry_symlink(e), "link");
 
 	/* mode */
@@ -436,7 +436,7 @@ DEFINE_TEST(test_entry)
 	archive_entry_fflags(e, &set, &clear);
 	assertEqualInt(UF_HIDDEN, set);
 	assertEqualInt(UF_NODUMP | UF_IMMUTABLE | UF_APPEND, clear);
-#elif defined(_WIN32) && !defined(CYGWIN)
+#elif defined(_WIN32) && !defined(__CYGWIN__)
 	archive_entry_copy_fflags_text_w(e, L"rdonly,hidden,nosystem");
 	archive_entry_fflags(e, &set, &clear);
 	assertEqualInt(FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_HIDDEN, set);
@@ -513,7 +513,6 @@ DEFINE_TEST(test_entry)
 	archive_entry_set_rdev(e, 532);
 	archive_entry_set_size(e, 987654321);
 	archive_entry_copy_sourcepath(e, "source");
-	archive_entry_set_symlink(e, "symlinkname");
 	archive_entry_set_uid(e, 83);
 	archive_entry_set_uname(e, "user");
 	/* Add an ACL entry. */
@@ -548,7 +547,7 @@ DEFINE_TEST(test_entry)
 	assertEqualInt(archive_entry_rdev(e2), 532);
 	assertEqualInt(archive_entry_size(e2), 987654321);
 	assertEqualString(archive_entry_sourcepath(e2), "source");
-	assertEqualString(archive_entry_symlink(e2), "symlinkname");
+	assertEqualString(archive_entry_symlink(e2), NULL);
 	assertEqualInt(archive_entry_uid(e2), 83);
 	assertEqualString(archive_entry_uname(e2), "user");
 
@@ -649,7 +648,7 @@ DEFINE_TEST(test_entry)
 	assertEqualInt(archive_entry_rdev(e2), 532);
 	assertEqualInt(archive_entry_size(e2), 987654321);
 	assertEqualString(archive_entry_sourcepath(e2), "source");
-	assertEqualString(archive_entry_symlink(e2), "symlinkname");
+	assertEqualString(archive_entry_symlink(e2), NULL);
 	assertEqualInt(archive_entry_uid(e2), 83);
 	assertEqualString(archive_entry_uname(e2), "user");
 
@@ -705,6 +704,13 @@ DEFINE_TEST(test_entry)
 	/* Release clone. */
 	archive_entry_free(e2);
 
+	/* Verify that symlink is copied over by `clone` */
+	archive_entry_set_symlink(e, "symlinkpath");
+	e2 = archive_entry_clone(e);
+	assertEqualString(archive_entry_hardlink(e2), NULL);
+	assertEqualString(archive_entry_symlink(e2), "symlinkpath");
+	archive_entry_free(e2);
+
 	/*
 	 * Test clear() implementation.
 	 */
@@ -874,6 +880,17 @@ DEFINE_TEST(test_entry)
 	if (pst == NULL)
 		return;
 	assertEqualInt(pst->st_uid, 22);
+
+	/* Check behavior with large sizes. */
+	archive_entry_set_size(e, INT64_MAX - 1);
+	assert((pst = archive_entry_stat(e)) != NULL);
+	if (pst == NULL)
+		return;
+	if (sizeof(pst->st_size) < sizeof(int64_t))
+		assertEqualInt(pst->st_size, 0);
+	else
+		assertEqualInt(pst->st_size, INT64_MAX - 1);
+
 	/* We don't need to check high-res fields here. */
 
 	/*
